@@ -3,7 +3,8 @@
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
     2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
-    2017, 2018, 2019 Massachusetts Institute of Technology
+    2017, 2018, 2019, 2020, 2021, 2022 Massachusetts Institute of
+    Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -391,7 +392,7 @@ negative numeric arg overrides kept-old-versions with minus the arg."
 			(let ()
 			  (let ((end (- nv total)))
 			    (do ((versions
-				  (list-tail
+				  (drop
 				   (sort (cdr file)
 					 (lambda (x y)
 					   (< (car x) (car y))))
@@ -798,7 +799,7 @@ Actions controlled by variables list-directory-brief-switches
 (define directory-listing-before-filename-regexp
   (let* ((l
 	  (char-set-union char-set:alphabetic
-			  (ascii-range->char-set #x80 #x100)))
+			  (ucs-range->char-set #x80 #x100)))
 	 (l? (rexp-optional l))
 	 (l-or-quote (char-set-union l (char-set #\')))
 	 (digit (string->char-set "0123456789"))
