@@ -3,7 +3,8 @@
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
     2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
-    2017, 2018 Massachusetts Institute of Technology
+    2017, 2018, 2019, 2020, 2021, 2022 Massachusetts Institute of
+    Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -397,10 +398,10 @@ Switches may be concatenated, e.g. `-lt' is equivalent to `-l -t'."
        (char-set-union (string->char-set "\"/:<>\\|")
 		       (string->char-set "*?"))))
   (set! char-set:valid-hpfs
-	(char-set-difference (ascii-range->char-set #x21 #x7F)
+	(char-set-difference (ucs-range->char-set #x21 #x7F)
 			     reserved-chars))
   (set! char-set:valid-windows-long
-	(char-set-difference (ascii-range->char-set #x20 #x100)
+	(char-set-difference (ucs-range->char-set #x20 #x100)
 			     reserved-chars)))
 
 (define char-set:valid-fat
