@@ -3,8 +3,8 @@
 # Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
 #     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
 #     2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014,
-#     2015, 2016, 2017, 2018, 2019 Massachusetts Institute of
-#     Technology
+#     2015, 2016, 2017, 2018, 2019, 2020, 2021, 2022 Massachusetts
+#     Institute of Technology
 #
 # This file is part of MIT/GNU Scheme.
 #
@@ -29,33 +29,36 @@ set -e
 
 . `dirname "${0}"`/functions.sh
 
-if [ ${#} -eq 2 ]; then
+if [ ${#} -eq 3 ]; then
     DIR=${1}
     EXE=${2}
+    ARCH=${3}
 else
-    echo "usage: ${0} <directory> <name>"
+    echo "usage: ${0} <directory> <name> <arch>"
     exit 1
 fi
 
-case ${EXE} in
-    mit-scheme-c|mit-scheme-native)
-	if test ! -f "${DIR}"/mit-scheme; then
-	    run_cmd rm -f "${DIR}"/mit-scheme
-	    run_cmd ln -s "${EXE}" "${DIR}"/mit-scheme
-	fi
-	;;
-    mit-scheme-*)
-	if test ! -f "${DIR}"/mit-scheme-native; then
-	    run_cmd rm -f "${DIR}"/mit-scheme-native
-	    run_cmd ln -s "${EXE}" "${DIR}"/mit-scheme-native
-	fi
-	if test ! -f "${DIR}"/mit-scheme; then
-	    run_cmd rm -f "${DIR}"/mit-scheme
-	    run_cmd ln -s mit-scheme-native "${DIR}"/mit-scheme
-	fi
-	;;
-esac
+link_name ()
+{
+    local TARGET=${DIR}/${1}
+    if ! test -e "${TARGET}"; then
+        run_cmd ln -s "${EXE}" "${TARGET}"
+    elif test -L "${TARGET}"; then
+        run_cmd rm -f "${TARGET}"
+        run_cmd ln -s "${EXE}" "${TARGET}"
+    fi
+}
 
-run_cmd rm -f "${DIR}"/scheme "${DIR}"/bchscheme
-run_cmd ln -s mit-scheme "${DIR}"/scheme
-run_cmd ln -s mit-scheme "${DIR}"/bchscheme
+unlink_name ()
+{
+    local TARGET=${DIR}/${1}
+    if test -L "${TARGET}"; then
+        run_cmd rm -f "${TARGET}"
+    fi
+}
+
+link_name mit-scheme-"${ARCH}"
+link_name mit-scheme
+link_name scheme
+unlink_name bchscheme
+unlink_name mit-scheme-native
