@@ -2,7 +2,8 @@
 
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
-    2006, 2007, 2008, 2009, 2010, 2011 Massachusetts Institute of
+    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
+    2017, 2018, 2019, 2020, 2021, 2022 Massachusetts Institute of
     Technology
 
 This file is part of MIT/GNU Scheme.
@@ -124,13 +125,14 @@ OS_channel_close (Tchannel channel)
 {
   if (! (CHANNEL_INTERNAL (channel)))
     {
-      if (0 > (UX_close (CHANNEL_DESCRIPTOR (channel))))
+      int status = (UX_close (CHANNEL_DESCRIPTOR (channel)));
+      MARK_CHANNEL_CLOSED (channel);
+      if (status < 0)
 	switch (errno)
 	  {
 	  case EINTR:	deliver_pending_interrupts ();			break;
 	  case EBADF:	error_system_call (errno, syscall_close);	break;
 	  }
-      MARK_CHANNEL_CLOSED (channel);
     }
 }
 
@@ -267,6 +269,16 @@ OS_channel_read (Tchannel channel, void * buffer, size_t nbytes)
 	if (errno == ERRNO_NONBLOCK)
 	  return (-1);
 #endif
+	/* From Emacs v25.2 src/process.c: On some OSs with ptys, when
+	   the process on one end of a pty exits, the other end gets
+	   an error reading with errno = EIO instead of getting an EOF
+	   (0 bytes read)...
+
+	   Thus, if a pty master and errno=EIO, return 0 bytes read. */
+	if (CHANNEL_TYPE (channel) == channel_type_unix_pty_master
+	    && errno == EIO)
+	  return (0);
+
 #endif /* not _ULTRIX */
 	  UX_prim_check_errno (syscall_read);
 	  continue;
@@ -537,6 +549,17 @@ OS_select_registry_length (select_registry_t registry)
   return (SR_N_FDS (r));
 }
 
+void
+OS_select_registry_entry (select_registry_t registry,
+			  unsigned int index,
+			  int * fd_r,
+			  unsigned int * mode_r)
+{
+  struct select_registry_s * r = registry;
+  (*fd_r) = ((SR_ENTRY (r, index)) -> fd);
+  (*mode_r) = (ENCODE_MODE ((SR_ENTRY (r, index)) -> events));
+}
+
 void
 OS_select_registry_result (select_registry_t registry,
 			   unsigned int index,
@@ -548,24 +571,63 @@ OS_select_registry_result (select_registry_t registry,
   (*mode_r) = (ENCODE_MODE ((SR_ENTRY (r, index)) -> revents));
 }
 
+static int
+safe_poll (struct pollfd *fds, nfds_t nfds, int blockp)
+{
+  int n;
+
+#ifdef HAVE_PPOLL
+  if (!blockp)
+    {
+      n = (UX_poll (fds, nfds, 0));
+    }
+  else
+    {
+      sigset_t old, new;
+
+      UX_sigfillset (&new);
+      UX_sigprocmask (SIG_SETMASK, &new, &old);
+      if ((OS_process_any_status_change ())
+	  || (pending_interrupts_p ()))
+	{
+	  errno = EINTR;
+	  n = -1;
+	}
+      else
+	{
+	  n = (UX_ppoll (fds, nfds, NULL, &old));
+	}
+      UX_sigprocmask (SIG_SETMASK, &old, NULL);
+    }
+#else /* not HAVE_PPOLL */
+  /* There is a signal "hole" here, but what can we do without ppoll()? */
+  INTERRUPTABLE_EXTENT
+    (n, (((OS_process_any_status_change ())
+	  || (pending_interrupts_p ()))
+	 ? ((errno = EINTR), (-1))
+	 : (UX_poll (fds, nfds, (blockp ? INFTIM : 0)))));
+#endif
+
+  return (n);
+}
+
 int
 OS_test_select_registry (select_registry_t registry, int blockp)
 {
   struct select_registry_s * r = registry;
   while (1)
     {
-      int nfds
-	= (poll ((SR_ENTRIES (r)),
-		 (SR_N_FDS (r)),
-		 (blockp ? INFTIM : 0)));
-      if (nfds >= 0)
+      int nfds = (safe_poll ((SR_ENTRIES (r)), (SR_N_FDS (r)), blockp));
+      if (nfds > 0)
 	return (nfds);
-      if (errno != EINTR)
+      if (nfds < 0 && errno != EINTR)
 	error_system_call (errno, syscall_select);
       if (OS_process_any_status_change ())
 	return (SELECT_PROCESS_STATUS_CHANGE);
       if (pending_interrupts_p ())
 	return (SELECT_INTERRUPT);
+      if (nfds == 0) /* and no status-change nor interrupts pending */
+	return (0);
     }
 }
 
@@ -577,17 +639,17 @@ OS_test_select_descriptor (int fd, int blockp, unsigned int mode)
   ((pfds [0]) . events) = (DECODE_MODE (mode));
   while (1)
     {
-      int nfds = (poll (pfds, 1, (blockp ? INFTIM : 0)));
+      int nfds = (safe_poll (pfds, 1, blockp));
       if (nfds > 0)
 	return (ENCODE_MODE ((pfds [0]) . revents));
-      if (nfds == 0)
-	return (0);
-      if (errno != EINTR)
+      if (nfds < 0 && errno != EINTR)
 	error_system_call (errno, syscall_select);
       if (OS_process_any_status_change ())
 	return (SELECT_PROCESS_STATUS_CHANGE);
       if (pending_interrupts_p ())
 	return (SELECT_INTERRUPT);
+      if (nfds == 0) /* and no status-change nor interrupts pending */
+	return (0);
     }
 }
 
@@ -705,34 +767,71 @@ OS_select_registry_result (select_registry_t registry,
     }
 }
 
+static int
+safe_select (int nfds, SELECT_TYPE *readfds, SELECT_TYPE *writefds, int blockp)
+{
+  int n;
+
+#ifdef HAVE_PSELECT
+  if (!blockp)
+    {
+      n = (UX_select (nfds, readfds, writefds, NULL, &zero_timeout));
+    }
+  else
+    {
+      sigset_t old, new;
+
+      UX_sigfillset (&new);
+      UX_sigprocmask (SIG_SETMASK, &new, &old);
+      if ((OS_process_any_status_change ())
+	  || (pending_interrupts_p ()))
+	{
+	  errno = EINTR;
+	  n = -1;
+	}
+      else
+	{
+	  n = (UX_pselect (nfds, readfds, writefds, NULL, NULL, &old));
+	}
+      UX_sigprocmask (SIG_SETMASK, &old, NULL);
+    }
+#else /* not HAVE_PSELECT */
+  /* There is a signal "hole" here, but what can we do without pselect()? */
+  INTERRUPTABLE_EXTENT
+    (n, (((OS_process_any_status_change ())
+	  || (pending_interrupts_p ()))
+	 ? ((errno = EINTR), (-1))
+	 : (UX_select (nfds, readfds, writefds, NULL,
+		       (blockp ? NULL : &zero_timeout)))));
+#endif
+
+  return (n);
+}
+
 int
 OS_test_select_registry (select_registry_t registry, int blockp)
 {
 #ifdef HAVE_SELECT
   struct select_registry_s * r = registry;
+
+  (* (SR_RREADERS (r))) = (* (SR_QREADERS (r)));
+  (* (SR_RWRITERS (r))) = (* (SR_QWRITERS (r)));
   while (1)
     {
-      int nfds;
-
-      (* (SR_RREADERS (r))) = (* (SR_QREADERS (r)));
-      (* (SR_RWRITERS (r))) = (* (SR_QWRITERS (r)));
-      INTERRUPTABLE_EXTENT
-	(nfds,
-	 ((OS_process_any_status_change ())
-	  ? ((errno = EINTR), (-1))
-	  : (UX_select (FD_SETSIZE,
-			(SR_RREADERS (r)),
-			(SR_RWRITERS (r)),
-			0,
-			(blockp ? 0 : (&zero_timeout))))));
-      if (nfds >= 0)
+      int nfds = (safe_select (FD_SETSIZE,
+			       (SR_RREADERS (r)),
+			       (SR_RWRITERS (r)),
+			       blockp));
+      if (nfds > 0)
 	return (nfds);
-      if (errno != EINTR)
+      if (nfds < 0 && errno != EINTR)
 	error_system_call (errno, syscall_select);
       if (OS_process_any_status_change ())
 	return (SELECT_PROCESS_STATUS_CHANGE);
       if (pending_interrupts_p ())
 	return (SELECT_INTERRUPT);
+      if (nfds == 0) /* and no status-change nor interrupts pending */
+	return (0);
     }
 #else
   error_system_call (ENOSYS, syscall_select);
@@ -744,41 +843,32 @@ int
 OS_test_select_descriptor (int fd, int blockp, unsigned int mode)
 {
 #ifdef HAVE_SELECT
+  SELECT_TYPE readable;
+  SELECT_TYPE writeable;
+
+  FD_ZERO (&readable);
+  if ((mode & SELECT_MODE_READ) != 0)
+    FD_SET (fd, (&readable));
+
+  FD_ZERO (&writeable);
+  if ((mode & SELECT_MODE_WRITE) != 0)
+    FD_SET (fd, (&writeable));
+
   while (1)
     {
-      SELECT_TYPE readable;
-      SELECT_TYPE writeable;
-      int nfds;
-
-      FD_ZERO (&readable);
-      if ((mode & SELECT_MODE_READ) != 0)
-	FD_SET (fd, (&readable));
-
-      FD_ZERO (&writeable);
-      if ((mode & SELECT_MODE_WRITE) != 0)
-	FD_SET (fd, (&writeable));
-
-      INTERRUPTABLE_EXTENT
-	(nfds,
-	 ((OS_process_any_status_change ())
-	  ? ((errno = EINTR), (-1))
-	  : (UX_select ((fd + 1),
-			(&readable),
-			(&writeable),
-			0,
-			(blockp ? 0 : (&zero_timeout))))));
+      int nfds = (safe_select (fd + 1, &readable, &writeable, blockp));
       if (nfds > 0)
 	return
 	  (((FD_ISSET (fd, (&readable))) ? SELECT_MODE_READ : 0)
 	   | ((FD_ISSET (fd, (&writeable))) ? SELECT_MODE_WRITE : 0));
-      if (nfds == 0)
-	return (0);
-      if (errno != EINTR)
+      if (nfds < 0 && errno != EINTR)
 	error_system_call (errno, syscall_select);
       if (OS_process_any_status_change ())
 	return (SELECT_PROCESS_STATUS_CHANGE);
       if (pending_interrupts_p ())
 	return (SELECT_INTERRUPT);
+      if (nfds == 0) /* and no status-change nor interrupts pending */
+	return (0);
     }
 #else
   error_system_call (ENOSYS, syscall_select);
@@ -787,3 +877,49 @@ OS_test_select_descriptor (int fd, int blockp, unsigned int mode)
 }
 
 #endif /* not HAVE_POLL */
+
+int
+OS_pause (bool blockp, bool ignore_status_change)
+{
+  int n;
+
+  if (!blockp)
+    {
+      if (!ignore_status_change && OS_process_any_status_change ())
+	return (SELECT_PROCESS_STATUS_CHANGE);
+      return (SELECT_INTERRUPT);
+    }
+
+#ifdef HAVE_SIGSUSPEND
+  {
+    sigset_t old, new;
+
+    UX_sigfillset (&new);
+    UX_sigprocmask (SIG_SETMASK, &new, &old);
+    if (!ignore_status_change && OS_process_any_status_change ())
+      n = SELECT_PROCESS_STATUS_CHANGE;
+    else if ((GET_INT_CODE) != 0)
+      n = SELECT_INTERRUPT;
+    else
+      {
+	UX_sigsuspend (&old);
+	if (OS_process_any_status_change ())
+	  n = SELECT_PROCESS_STATUS_CHANGE;
+	else
+	  n = SELECT_INTERRUPT;
+      }
+    UX_sigprocmask (SIG_SETMASK, &old, NULL);
+  }
+#else /* not HAVE_SIGSUSPEND */
+  INTERRUPTABLE_EXTENT
+    (n, (((!ignore_status_change && (OS_process_any_status_change ()))
+	  || ((GET_INT_CODE) != 0))
+	 ? ((errno = EINTR), (-1))
+	 : ((UX_pause ()), (0))));
+  if (OS_process_any_status_change())
+    n = SELECT_PROCESS_STATUS_CHANGE;
+  else
+    n = SELECT_INTERRUPT;
+#endif
+  return (n);
+}
